//
// Created by philipp on 12.02.18.
//

#include "../../include/mainNode/PlatoonController.h"

namespace car
{

PlatoonController::PlatoonController(ros::NodeHandle & nh,
                                     platoonProtocol::VehicleFacade & c2c,
                                     pc2car::CommandReceiver::Ptr pc,
                                     EgoMotion & egoMotion)
    : messageOStream(nh, "PlatoonController")
      , cruiseControllerNotify()
      , c2c(c2c)
      , pc(std::move(pc))
      , egoMotion(egoMotion)
      , platoonConfig(c2c.getPlatoonConfig())
{}

void PlatoonController::updatePcConfig()
{
    auto otherPS = pc->getPlatoonSpeed();
    auto otherIPD = pc->getInnerPlatoonDistance();

    if (otherPS.getTimestamp() > PS.getTimestamp())
    { PS = otherPS; }
    if (otherIPD.getTimestamp() > IPD.getTimestamp())
    { IPD = otherIPD; }

    cruiseControllerNotify();
}

void PlatoonController::updateDesSpeed()
{
    auto otherDesSpeed = pc->getInnerPlatoonDistance();
    if (otherDesSpeed.getTimestamp() > desSpeed.getTimestamp())
    { desSpeed = otherDesSpeed; }

    cruiseControllerNotify();
}

void PlatoonController::setupC2C()
{
    bool wantsPlatoon = pc->isPlatoonEnabled().get();

    if (c2cAlive)
    { return; } // if C2C is already alive ... do nothing
    if (!wantsPlatoon)
    { return; } // if we dont want platoon ...

    // get role according to distance
    if (egoMotion.getDistance() == std::numeric_limits<float>::infinity())
    {
        c2c.reset(platoonProtocol::Vehicle::Role::LEADER);
        c2c.setInnerPlatoonDistance(IPD.get());
        c2c.setPlatoonSpeed(PS.get());
    }
    else
    {
        c2c.reset(platoonProtocol::Vehicle::Role::FOLLOWER);
    }
    c2c.createPlatoon();
    c2cAlive = true;
}

void PlatoonController::run()
{
    switch (currState.load())
    {
        case PlatoonState::ACC:
        {
            run_ACC();
            break;
        }
        case PlatoonState::CACC_FV:
        {
            run_CACC_FV();
            break;
        }
        case PlatoonState::CACC_LV:
        {
            run_CACC_LV();
            break;
        }
    }
}

void PlatoonController::run_ACC()
{
    messageOStream.write("run - ACC", "");
    bool wantsPlatoon = pc->isPlatoonEnabled().get();

    if (!wantsPlatoon)
    {
        if (c2cAlive)
        {
            c2c.leavePlatoon();
            c2cAlive = false;
        }
        updateDesSpeed();
        return;
    }

    // wantsPlatoon
    if (!c2cAlive)
    {
        updateDesSpeed();
        setupC2C();
        return;
    }

    // c2cAlive + wantsPlatoon
    if (c2c.isPlatoonRunning())
    {
        if (c2c.getRole() == platoonProtocol::Vehicle::Role::LEADER)
        {
            updatePcConfig();
            currState = PlatoonState::CACC_LV;
        }
        else
        { currState = PlatoonState::CACC_FV; }
        return;
    }

    // !inPlatoon + c2cAlive + wantsPlatoon
    bool isLeader = c2c.getRole() == platoonProtocol::Vehicle::Role::LEADER;
    bool hasFiniteDistance = egoMotion.getDistance() < std::numeric_limits<float>::infinity();
    if (isLeader == inPlatoonRange)
    { // role does not fit
        c2c.leavePlatoon();
        c2cAlive = false;
        updateDesSpeed();
        setupC2C();
        return;
    }

    // role fits + !inPlatoon + c2cAlive + wantsPlatoon
    if (c2c.getRole() == platoonProtocol::Vehicle::Role::LEADER)
    { updatePcConfig(); }
    updateDesSpeed();
}

void PlatoonController::run_CACC_FV()
{
    messageOStream.write("run - FV", "");
    bool inPlatoon = c2c.isPlatoonRunning();
    bool wantsPlatoon = pc->isPlatoonEnabled().get();

    // Although this value will not be used in this method it still
    // needs to be updated, so the new value can be pulled be CC
    platoonConfig = c2c.getPlatoonConfig();

    messageOStream << MessageOStream::key << "run_CACC_FV"
                   << MessageOStream::value << "inPlatoon: " << inPlatoon << "\nwantsPlatoon: " << wantsPlatoon
                   << MessageOStream::flush;

    if (inPlatoon && wantsPlatoon)
    {
        cruiseControllerNotify();
        return;
    }

    if (inPlatoon && !wantsPlatoon)
    { c2c.leavePlatoon(); }

    c2cAlive = false;
    currState = PlatoonState::ACC;
    cruiseControllerNotify();
    run_ACC();
}

void PlatoonController::run_CACC_LV()
{
    messageOStream.write("run - LV", "");
    bool inPlatoon = c2c.isPlatoonRunning();
    bool wantsPlatoon = pc->isPlatoonEnabled().get();

    messageOStream << MessageOStream::key << "run_CACC_LV"
                   << MessageOStream::value << "inPlatoon: " << inPlatoon << "\nwantsPlatoon: " << wantsPlatoon
                   << MessageOStream::flush;

    if (inPlatoon && wantsPlatoon)
    {
        updatePcConfig();
        c2c.setInnerPlatoonDistance(IPD.get());
        return;
    }

    if (inPlatoon && !wantsPlatoon)
    {
        c2c.leavePlatoon();
    }

    c2cAlive = false;
    currState = PlatoonState::ACC;
    cruiseControllerNotify();
    run_ACC();
}

}
